@@ -15,6 +15,7 @@
 #include <sys/wait.h>
 #endif
 #include <signal.h>
+#include <strings.h>
 
 #ifdef	__WATT32__
 #define	read(a,b,c)	read_s(a,b,c)
@@ -2084,7 +2085,7 @@ base64_encode(const char *src, size_t len)
 	}
 	Strcatc(dest, '=');
     }
-    Strnulterm(dest);
+    dest->ptr[dest->length]='\0';
     return dest;
 #undef Strcatc
 }
