@@ -1,4 +1,5 @@
 /* vi: set sw=4 ts=8 ai sm noet : */
+#include "buffer.h"
 #include "fm.h"
 #include "local.h"
 #include "myctype.h"
@@ -22,6 +23,8 @@ extern int do_getch(void);
 
 static Str strBuf;
 static Lineprop strProp[STR_LEN];
+static Str ynkBuf;
+static int ynkCon;
 
 static Str CompleteBuf;
 static Str CFileName;
@@ -32,18 +35,41 @@ static char **CFileBuf = NULL;
 static int NCFileBuf;
 static int NCFileOffset;
 
-static void insertself(char c),
-_mvR(void), _mvL(void), _mvRw(void), _mvLw(void), delC(void), insC(void),
-_mvB(void), _mvE(void), _enter(void), _quo(void), _bs(void), _bsw(void),
-killn(void), killb(void), _inbrk(void), _esc(void), _editor(void),
-_prev(void), _next(void), _compl(void), _tcompl(void),
-_dcompl(void), _rdcompl(void), _rcompl(void);
+static void _bs(void);
+static void _bsw(void);
+static void _compl(void);
+static void _cy(void);
+static void _dcompl(void);
+static void _editor(void);
+static void _enter(void);
+static void _esc(void);
+static void _inbrk(void);
+static void _isrch(void);
+static void _iword(void);
+static void _mvB(void);
+static void _mvE(void);
+static void _mvL(void);
+static void _mvLw(void);
+static void _mvR(void);
+static void _mvRw(void);
+static void _next(void);
+static void _paste(void);
+static void _prev(void);
+static void _quo(void);
+static void _rcompl(void);
+static void _rdcompl(void);
+static void _tcompl(void);
+static void delC(void);
+static void insC(void);
+static void killb(void);
+static void killn(void);
+static void _noop(void);
+
 #ifdef __EMX__
 static int getcntrl(void);
 #endif
 
 static int terminated(unsigned char c);
-#define iself ((void(*)(void))insertself)
 
 static void next_compl(int next);
 static void next_dcompl(int next);
@@ -54,11 +80,11 @@ void (*InputKeymap[32]) (void) = {
 /*  C-@     C-a     C-b     C-c     C-d     C-e     C-f     C-g     */
     _compl, _mvB,   _mvL,   _inbrk, delC,   _mvE,   _mvR,   _inbrk,
 /*  C-h     C-i     C-j     C-k     C-l     C-m     C-n     C-o     */
-    _bs,    iself,  _enter, killn,  iself,  _enter, _next,  _editor,
+    _bs,    _noop,  _enter, killn,  _noop,  _enter, _next,  _editor,
 /*  C-p     C-q     C-r     C-s     C-t     C-u     C-v     C-w     */
-    _prev,  _quo,   _bsw,   iself,  _mvLw,  killb,  _quo,   _bsw,
+    _prev,  _quo,   _bsw,   _noop,  _mvLw,  killb,  _quo,   _bsw,
 /*  C-x     C-y     C-z     C-[     C-\     C-]     C-^     C-_     */
-    _tcompl,_mvRw,  iself,  _esc,   iself,  iself,  iself,  iself,
+    _tcompl,_cy,    _noop,  _esc,   _isrch, _iword, _noop,  _noop,
 };
 /* *INDENT-ON* */
 
@@ -70,11 +96,9 @@ static int CPos, CLen, offset;
 static int i_cont, i_broken, i_quote;
 static int cm_mode, cm_next, cm_clear, cm_disp_next, cm_disp_clear;
 static int need_redraw, is_passwd;
-static int move_word;
 
 static Hist *CurrentHist;
 static Str strCurrentBuf;
-static int use_hist;
 #ifdef USE_M17N
 static void ins_char(Str str);
 #endif
@@ -92,16 +116,10 @@ inputLineHistSearch(const char *prompt, const char *def_str,
 #endif
 
     is_passwd = FALSE;
-    move_word = TRUE;
 
-    CurrentHist = hist;
-    if (hist != NULL) {
-	use_hist = TRUE;
+    if ((CurrentHist = hist))
 	strCurrentBuf = NULL;
-    }
-    else {
-	use_hist = FALSE;
-    }
+
     if (flag & IN_URL) {
 	cm_mode = CPL_ALWAYS | CPL_URL;
     }
@@ -111,7 +129,6 @@ inputLineHistSearch(const char *prompt, const char *def_str,
     else if (flag & IN_PASSWORD) {
 	cm_mode = CPL_NEVER;
 	is_passwd = TRUE;
-	move_word = FALSE;
     }
     else if (flag & IN_COMMAND)
 	cm_mode = CPL_ON;
@@ -183,7 +200,8 @@ inputLineHistSearch(const char *prompt, const char *def_str,
 	cm_clear = TRUE;
 	cm_disp_clear = TRUE;
 	if (!i_quote &&
-	    (((cm_mode & CPL_ALWAYS) && (c == CTRL_I || (space_autocomplete && c == ' '))) ||
+	    (((cm_mode & CPL_ALWAYS) &&
+	      (c == CTRL_I || (space_autocomplete && c == ' '))) ||
 	     ((cm_mode & CPL_ON) && (c == CTRL_I)))) {
 	    if (emacs_like_lineedit && cm_next) {
 		_dcompl();
@@ -229,6 +247,7 @@ inputLineHistSearch(const char *prompt, const char *def_str,
 	    cm_disp_next = -1;
 	    if (CLen + tmp->length > STR_LEN || !tmp->length)
 		goto next_char;
+	    ynkCon = 0;
 	    ins_char(tmp);
 	    if (incrfunc)
 		incrfunc(-1, strBuf, strProp);
@@ -240,6 +259,7 @@ inputLineHistSearch(const char *prompt, const char *def_str,
 	    cm_disp_next = -1;
 	    if (CLen >= STR_LEN)
 		goto next_char;
+	    ynkCon = 0;
 	    insC();
 	    strBuf->ptr[CPos] = c;
 	    if (!is_passwd && get_mctype(&c) == PC_CTRL)
@@ -273,7 +293,7 @@ inputLineHistSearch(const char *prompt, const char *def_str,
     if (flag & (IN_FILENAME | IN_COMMAND)) {
 	SKIP_BLANKS(p);
     }
-    if (use_hist && !(flag & IN_URL) && *p != '\0') {
+    if (hist && !(flag & IN_URL) && *p != '\0') {
 	char *q = lastHist(hist);
 	if (!q || strcmp(q, p))
 	    pushHist(hist, p);
@@ -375,10 +395,10 @@ addStr(char *p, Lineprop *pr, int len, int offset, int limit)
     }
 }
 
-#ifdef USE_M17N
 static void
 ins_char(Str str)
 {
+#ifdef USE_M17N
     const char *p = str->ptr, *ep = p + str->length;
     Lineprop ctype;
     int len;
@@ -408,8 +428,22 @@ ins_char(Str str)
 	    }
 	}
     }
-}
+
+#else
+    char c;
+    int i;
+
+    for (i = 0, c = str->ptr[i]; c; c = str->ptr[++i]) {
+	    insC();
+	    strBuf->ptr[CPos] = c;
+	    if (!is_passwd && get_mctype(&c) == PC_CTRL)
+		strProp[CPos] = PC_CTRL;
+	    else
+		strProp[CPos] = PC_ASCII;
+	    CPos++;
+    }
 #endif
+}
 
 static void
 _esc(void)
@@ -419,7 +453,7 @@ _esc(void)
     switch (c = getch()) {
     case '[':
     case 'O':
-	switch (c = getch()) {
+	switch (getch()) {
 	case 'A':
 	    _prev();
 	    break;
@@ -484,22 +518,46 @@ insC(void)
 static void
 delC(void)
 {
-    int i = CPos;
     int delta = 1;
 
     if (CLen == CPos)
 	return;
 #ifdef USE_M17N
-    while (i + delta < CLen && strProp[i + delta] & PC_WCHAR2)
+    while (CPos + delta < CLen && strProp[CPos + delta] & PC_WCHAR2)
 	delta++;
 #endif
-    for (i = CPos; i < CLen; i++) {
+    for (int i = CPos; i < CLen; i++)
 	strProp[i] = strProp[i + delta];
+
+    if (!is_passwd) {
+	if (!ynkBuf)
+	    ynkBuf = Strnew();
+	if (!ynkCon)
+	    Strshrink(ynkBuf, ynkBuf->length); /* TODO(rkta): same as clear? */
+	Strinsert_charp_n(ynkBuf, 0, &strBuf->ptr[CPos], delta);
+	ynkCon = 1;
     }
     Strdelete(strBuf, CPos, delta);
     CLen -= delta;
 }
 
+static void
+_cy(void)
+{
+    if (rl_paste)
+	_paste();
+    else
+	_mvRw();
+}
+
+static void
+_paste(void)
+{
+    if (!ynkBuf) return;
+    ins_char(ynkBuf);
+    ynkCon = 0;
+}
+
 static void
 _mvL(void)
 {
@@ -509,6 +567,7 @@ _mvL(void)
     while (CPos > 0 && strProp[CPos] & PC_WCHAR2)
 	CPos--;
 #endif
+    ynkCon = 0;
 }
 
 static void
@@ -522,9 +581,10 @@ _mvLw(void)
 	if (CPos > 0 && strProp[CPos] & PC_WCHAR2)
 	    CPos--;
 #endif
-	if (!move_word)
+	if (is_passwd)
 	    break;
     }
+    ynkCon = 0;
 }
 
 static void
@@ -538,9 +598,10 @@ _mvRw(void)
 	if (CPos < CLen && strProp[CPos] & PC_WCHAR2)
 	    CPos++;
 #endif
-	if (!move_word)
+	if (is_passwd)
 	    break;
     }
+    ynkCon = 0;
 }
 
 static void
@@ -552,13 +613,18 @@ _mvR(void)
     while (CPos < CLen && strProp[CPos] & PC_WCHAR2)
 	CPos++;
 #endif
+    ynkCon = 0;
 }
 
 static void
 _bs(void)
 {
+    int y;
+
     if (CPos > 0) {
+	y = ynkCon;
 	_mvL();
+	ynkCon = y;
 	delC();
     }
 }
@@ -566,10 +632,13 @@ _bs(void)
 static void
 _bsw(void)
 {
-    int t = 0;
+    int t = 0, y;
+
     while (CPos > 0 && !t) {
+	y = ynkCon;
 	_mvL();
-	t = (move_word && terminated(strBuf->ptr[CPos - 1]));
+	ynkCon = y;
+	t = (!is_passwd && terminated(strBuf->ptr[CPos - 1]));
 	delC();
     }
 }
@@ -581,14 +650,23 @@ _enter(void)
 }
 
 static void
-insertself(const char c)
+_iword(void)
 {
-    if (CLen >= STR_LEN)
-	return;
-    insC();
-    strBuf->ptr[CPos] = c;
-    strProp[CPos] = (is_passwd) ? PC_ASCII : PC_CTRL;
-    CPos++;
+    ins_char(Strnew_charp(GetWord(Currentbuf)));
+}
+
+static void
+_noop(void)
+{
+    return;
+}
+
+extern const char *SearchString; /* TODO(rkta): No forward decl here, but menu.c has it also as static. */
+static void
+_isrch(void)
+{
+    ins_char(Strnew_charp(SearchString));
+    ynkCon = 0;
 }
 
 static void
@@ -601,12 +679,14 @@ static void
 _mvB(void)
 {
     CPos = 0;
+    ynkCon = 0;
 }
 
 static void
 _mvE(void)
 {
     CPos = CLen;
+    ynkCon = 0;
 }
 
 static void
@@ -1010,7 +1090,7 @@ _prev(void)
     Hist *hist = CurrentHist;
     char *p;
 
-    if (!use_hist)
+    if (!hist)
 	return;
     if (strCurrentBuf) {
 	p = prevHist(hist);
@@ -1036,7 +1116,7 @@ _next(void)
     Hist *hist = CurrentHist;
     char *p;
 
-    if (!use_hist)
+    if (!hist)
 	return;
     if (strCurrentBuf == NULL)
 	return;
